@@ -35,7 +35,9 @@
 
 #include <glib.h>
 #include <glib/gi18n-lib.h>
+#include <glib/gstdio.h>
 
+#include <errno.h>
 #include <string.h>
 
 #define LOC_DELIMITER   ":"
@@ -68,6 +70,7 @@ typedef struct _GkrLocationManagerPrivate GkrLocationManagerPrivate;
 struct _GkrLocationManagerPrivate {
 #ifdef WITH_HAL
 	LibHalContext *hal_ctx;
+	gboolean hal_inited;
 	guint hal_retry;
 	DBusConnection *dbus_connection;
 #endif
@@ -320,8 +323,10 @@ hal_device_property (LibHalContext *hal_ctx, const char *udi, const char *key,
 		
 		g_message ("adding removable location: %s at %s", name, mount); 
 		gkr_location_manager_register (locmgr, name, mount, friendly);
+		
 		locvol = g_hash_table_lookup (pv->volumes_by_name, name);
-		locvol->hal_volume = TRUE;
+		if (locvol)
+			locvol->hal_volume = TRUE;
 
 	/* A mount was removed? */		
 	} else if (!is_mounted && g_hash_table_lookup (pv->volumes_by_name, name)) {
@@ -395,9 +400,14 @@ location_manager_hal_uninit (GkrLocationManager *locmgr)
 
 	if (pv->hal_ctx) {
 		dbus_error_init (&error);
-		if (pv->dbus_connection != NULL && !libhal_ctx_shutdown (pv->hal_ctx, &error)) {
-			g_warning ("failed to shutdown HAL context: %s\n", error.message);
-			dbus_error_free (&error);
+		if (pv->dbus_connection != NULL) {
+			if (pv->hal_inited) {
+				if (libhal_ctx_shutdown (pv->hal_ctx, &error)) {
+					g_warning ("failed to shutdown HAL context: %s\n", error.message);
+					dbus_error_free (&error);
+				}
+				pv->hal_inited = FALSE;
+			}
 		} 
 		
 		if (!libhal_ctx_free (pv->hal_ctx)) 
@@ -489,6 +499,7 @@ location_manager_hal_init (GkrLocationManager *locmgr)
 		goto failed;
 	}
 	
+	pv->hal_inited = TRUE;
 	libhal_ctx_set_user_data (pv->hal_ctx, locmgr);
 	
 	populate_all_volumes (locmgr);
@@ -1014,6 +1025,27 @@ gkr_location_to_display (GQuark loc)
 	return display;
 }
 
+gboolean
+gkr_location_is_volume (GQuark loc)
+{
+	const gchar *sloc;
+	const gchar *delim;
+	
+	if (!loc)
+		return FALSE;
+	
+	sloc = g_quark_to_string (loc);
+	g_return_val_if_fail (sloc, FALSE);
+
+	delim = strchr (sloc, LOC_DELIMITER_C);
+	if (!delim) {
+		g_warning ("The '%s' location is invalid", sloc);
+		return FALSE;
+	}
+	
+	return (delim[1] == 0);
+}
+
 gboolean
 gkr_location_is_descendant (GQuark parent, GQuark descendant)
 {
@@ -1101,3 +1133,32 @@ gkr_location_write_file (GQuark loc, const guchar *data, gssize len, GError **er
 	
 	return ret;
 }
+
+gboolean
+gkr_location_delete_file (GQuark loc, GError **err)
+{
+	gchar *path;
+	int eno;
+	
+	g_return_val_if_fail (loc != 0, FALSE);
+	g_return_val_if_fail (!err || !*err, FALSE);
+
+	/* Should be successful when file doesn't exist */
+	path = gkr_location_to_path (loc);
+	if (!path)
+		return TRUE;
+
+	if (g_unlink (path) < 0) {
+		eno = errno;
+		
+		/* Should be successful when file doesn't exist */
+		if (eno != ENOENT) {
+			g_set_error (err, G_FILE_ERROR, g_file_error_from_errno (eno), 
+			             _("Couldn't delete the file: %s"), g_strerror (eno));
+			return FALSE;
+		}
+	}
+	
+	g_free (path);
+	return TRUE;	
+}
