@@ -25,12 +25,16 @@
 #include "gkr-daemon.h"
 
 #include "common/gkr-buffer.h"
+#include "common/gkr-daemon-util.h"
 #include "common/gkr-location.h"
+#include "common/gkr-secure-memory.h"
 
+#include "keyrings/gkr-keyring.h"
+#include "keyrings/gkr-keyring-item.h"
 #include "keyrings/gkr-keyrings.h"
+#include "keyrings/gkr-keyring-login.h"
 
 #include "library/gnome-keyring.h"
-#include "library/gnome-keyring-memory.h"
 #include "library/gnome-keyring-private.h"
 #include "library/gnome-keyring-proto.h"
 
@@ -53,98 +57,6 @@
 /* for requesting list access to items */
 #define  GNOME_KEYRING_ACCESS_LIST 0
 
-static void
-save_keyring_password_in_login (GkrKeyring *keyring, const gchar *password)
-{
-	GkrKeyring *login;
-	GnomeKeyringAttributeList *attrs;
-	GkrKeyringItem *item;
-	
-	login = gkr_keyrings_get_login ();
-	if (!login || login->locked)
-		return;
-		
-	if (!keyring->location)
-		return;
-		
-	attrs = gnome_keyring_attribute_list_new ();
-	gnome_keyring_attribute_list_append_string (attrs, "keyring", 
-	                                            gkr_location_to_string (keyring->location));
-		
-	item = gkr_keyring_find_item (login, GNOME_KEYRING_ITEM_CHAINED_KEYRING_PASSWORD, attrs);
-	
-	if (!item) {
-		item = gkr_keyring_item_create (login, GNOME_KEYRING_ITEM_CHAINED_KEYRING_PASSWORD);
-		gkr_keyring_add_item (login, item);
-	}
-	
-	g_free (item->display_name);
-	/* TRANSLATORS: this is the title for an item */
-	item->display_name = g_strdup_printf (_("Unlock password for %s keyring"), 
-	                                      keyring->keyring_name);
-	gnome_keyring_free_password (item->secret);
-	item->secret = gnome_keyring_memory_strdup (password);
-	
-	gnome_keyring_attribute_list_free (item->attributes);
-	item->attributes = attrs;
-	
-	gkr_keyring_save_to_disk (login);
-}
-
-static const gchar*
-lookup_keyring_password_in_login (GkrKeyring *keyring)
-{
-	GkrKeyring *login;
-	GkrKeyringItem *item;
-	GnomeKeyringAttributeList *attrs;
-	
-	login = gkr_keyrings_get_login ();
-	if (!login || login->locked)
-		return NULL;
-
-	if (!keyring->location)
-		return NULL;
-				
-	attrs = gnome_keyring_attribute_list_new ();
-	gnome_keyring_attribute_list_append_string (attrs, "keyring", 
-	                                            gkr_location_to_string (keyring->location));
-		
-	item = gkr_keyring_find_item (login, GNOME_KEYRING_ITEM_CHAINED_KEYRING_PASSWORD, attrs);
-	gnome_keyring_attribute_list_free (attrs);
-	
-	if (item)
-		return item->secret;
-		
-	return NULL;
-}
-
-static void
-remove_keyring_password_from_login (GkrKeyring *keyring)
-{
-	GkrKeyring *login;
-	GkrKeyringItem *item;
-	GnomeKeyringAttributeList *attrs;
-	
-	login = gkr_keyrings_get_login ();
-	if (!login || login->locked)
-		return;
-		
-	if (!keyring->location)
-		return;
-
-	attrs = gnome_keyring_attribute_list_new ();
-	gnome_keyring_attribute_list_append_string (attrs, "keyring", 
-	                                            gkr_location_to_string (keyring->location));
-		
-	item = gkr_keyring_find_item (login, GNOME_KEYRING_ITEM_CHAINED_KEYRING_PASSWORD, attrs);
-	gnome_keyring_attribute_list_free (attrs);
-	
-	if (item) {
-		gkr_keyring_remove_item (login, item);
-		gkr_keyring_save_to_disk (login);
-	}	
-}
-
 static gboolean
 app_ref_match (GnomeKeyringApplicationRef *app1,
 	       GnomeKeyringApplicationRef *app2)
@@ -270,6 +182,7 @@ check_acl_ask_request (GkrAskRequest* ask, GnomeKeyringApplicationRef *app)
 		return GKR_ASK_STOP_REQUEST;
 	}
 	
+#ifdef ENABLE_ACL_PROMPTS
 	/* Did prompting already occur? */
 	if (ask->response) {
 		
@@ -285,6 +198,10 @@ check_acl_ask_request (GkrAskRequest* ask, GnomeKeyringApplicationRef *app)
 	
 	/* Continue with prompting */
 	return GKR_ASK_DONT_CARE;
+#else /* !ENABLE_ACL_PROMPTS */
+	ask->response = GKR_ASK_RESPONSE_ALLOW;
+	return GKR_ASK_STOP_REQUEST;
+#endif /* ENABLE_ACL_PROMPTS */
 }
 
 static gboolean
@@ -369,63 +286,6 @@ request_item_access (GkrKeyringRequest *req, GkrKeyringItem *item,
 	return ret;
 }
 
-static gboolean 
-check_keyring_ask_request (GkrAskRequest* ask)
-{
-	GkrKeyring *keyring;
-	const gchar *password;
-	
-	keyring = GKR_KEYRING (gkr_ask_request_get_object (ask));
-	g_assert (GKR_IS_KEYRING (keyring));
-
-	if (!keyring->locked) {
-		ask->response = GKR_ASK_RESPONSE_ALLOW;
-		return GKR_ASK_STOP_REQUEST;
-	}
-	
-	/* If they typed a password, try it out */
-	if (ask->response >= GKR_ASK_RESPONSE_ALLOW) {
-		
-		g_assert (ask->typed_password);
-		if (!gkr_keyring_unlock (keyring, ask->typed_password)) {
-			/* Bad password, try again */
-			ask->response = GKR_ASK_RESPONSE_NONE;
-			return GKR_ASK_CONTINUE_REQUEST;
-		}
-		
-		/* Did they ask us to remember the password? */
-		if (ask->checked)
-			save_keyring_password_in_login (keyring, ask->typed_password);
-	}
-	
-	/* 
-	 * We can automatically unlock keyrings that have their password
-	 * stored in the 'login' keyring.
-	 */
-	password = lookup_keyring_password_in_login (keyring);
-	if (password) {
-		if (gkr_keyring_unlock (keyring, password)) {
-			
-			/* A good password, unlocked, all done */
-			ask->response = GKR_ASK_RESPONSE_ALLOW;
-			return GKR_ASK_STOP_REQUEST;
-			
-		} else {
-			
-			/* A bad internal password */
-			remove_keyring_password_from_login (keyring);
-		}
-	}	
-
-	/* If the keyring is unlocked then no need to continue */
-	if (!keyring->locked) {
-		ask->response = GKR_ASK_RESPONSE_ALLOW;
-		return GKR_ASK_STOP_REQUEST;
-	}
-	
-	return GKR_ASK_DONT_CARE;
-}
-
 static gboolean
 request_keyring_access (GkrKeyringRequest *req, GkrKeyring *keyring)
 {
@@ -503,11 +363,11 @@ request_keyring_access (GkrKeyringRequest *req, GkrKeyring *keyring)
 	 * to unlock automatically next time. 
 	 */
 	login = gkr_keyrings_get_login ();
-	if (login && login != keyring)
+	if (login != keyring && gkr_keyring_login_is_usable ())
 		gkr_ask_request_set_check_option (ask, _("Automatically unlock this keyring when I log in."));
 	
 	/* Intercept item access requests to see if we still need to prompt */
-	g_signal_connect (ask, "check-request", G_CALLBACK (check_keyring_ask_request), NULL);
+	g_signal_connect (ask, "check-request", G_CALLBACK (gkr_keyring_ask_check_unlock), NULL);
 	
 	g_free (primary);
 	g_free (message);
@@ -522,7 +382,7 @@ request_keyring_access (GkrKeyringRequest *req, GkrKeyring *keyring)
 
 static gboolean
 request_new_keyring_password (GkrKeyringRequest *req, const char *keyring_name, 
-                              gchar **password, GQuark *base_loc)
+                              gchar **password, GQuark *volume)
 {
 	GnomeKeyringApplicationRef *app = req->app_ref;
 	GkrAskRequest* ask;
@@ -594,8 +454,8 @@ request_new_keyring_password (GkrKeyringRequest *req, const char *keyring_name,
 	ret = ask->response >= GKR_ASK_RESPONSE_ALLOW;
 	if (ret) {
 		g_free (*password);
-		*password = gnome_keyring_memory_strdup (ask->typed_password);
-		*base_loc = ask->location_selected;
+		*password = gkr_secure_strdup (ask->typed_password);
+		*volume = ask->location_selected;
 	}
 	
 	g_object_unref (ask);
@@ -690,10 +550,10 @@ request_change_keyring_password (GkrKeyringRequest *req, GkrKeyring* keyring,
 	ret = ask->response >= GKR_ASK_RESPONSE_ALLOW;
 	if (ret) {
 		g_free (*password);
-		*password = gnome_keyring_memory_strdup (ask->typed_password);
+		*password = gkr_secure_strdup (ask->typed_password);
 		
 		g_free (*original);
-		*original = gnome_keyring_memory_strdup (ask->original_password);
+		*original = gkr_secure_strdup (ask->original_password);
 	}
 	
 	g_object_unref (ask);
@@ -717,7 +577,7 @@ check_keyring_default_request (GkrAskRequest* ask)
 		g_assert (ask->typed_password);
 		
 		/* Create the new keyring */
-		keyring = gkr_keyring_create (GKR_LOCATION_BASE_LOCAL, "default", 
+		keyring = gkr_keyring_create (GKR_LOCATION_VOLUME_LOCAL, "default", 
 		                              ask->typed_password);
 		if (keyring == NULL) {
 			g_warning ("couldn't create default keyring");
@@ -823,7 +683,7 @@ change_keyring_password (GkrKeyring *keyring,  const char *password)
 	if (keyring->locked) {
 		return GNOME_KEYRING_RESULT_DENIED;
 	} else { 
-		keyring->password = gnome_keyring_memory_strdup (password);
+		keyring->password = gkr_secure_strdup (password);
 		gkr_keyring_save_to_disk (keyring);
 		return GNOME_KEYRING_RESULT_OK;
 	}
@@ -1004,7 +864,7 @@ static gboolean
 op_create_keyring (GkrBuffer *packet, GkrBuffer *result,
                    GkrKeyringRequest *req)
 {
-	GQuark base_loc = GKR_LOCATION_BASE_LOCAL;
+	GQuark volume = GKR_LOCATION_VOLUME_LOCAL;
 	char *keyring_name, *password;
 	GkrKeyring *keyring;
 	GnomeKeyringOpCode opcode;
@@ -1029,12 +889,12 @@ op_create_keyring (GkrBuffer *packet, GkrBuffer *result,
 	}
 	
 	/* Let user pick password if necessary*/
-	if (!request_new_keyring_password (req, keyring_name, &password, &base_loc)) {
+	if (!request_new_keyring_password (req, keyring_name, &password, &volume)) {
 		gkr_buffer_add_uint32 (result, GNOME_KEYRING_RESULT_DENIED);
 		goto out;
 	}
 	
-	keyring = gkr_keyring_create (base_loc, keyring_name, password);
+	keyring = gkr_keyring_create (volume, keyring_name, password);
 	if (keyring == NULL) {
 		gkr_buffer_add_uint32 (result, GNOME_KEYRING_RESULT_DENIED);
 		goto out;
@@ -1051,7 +911,7 @@ op_create_keyring (GkrBuffer *packet, GkrBuffer *result,
 	
  out:
 	g_free (keyring_name);
-	gnome_keyring_free_password (password);
+	gkr_secure_strfree (password);
 
 	return TRUE;
 }
@@ -1098,7 +958,7 @@ op_unlock_keyring (GkrBuffer *packet, GkrBuffer *result,
 
  out:
 	g_free (keyring_name);
-	gnome_keyring_free_password (password);
+	gkr_secure_strfree (password);
 
 	return TRUE;
 }
@@ -1109,7 +969,7 @@ op_delete_keyring (GkrBuffer *packet, GkrBuffer *result,
                    GkrKeyringRequest *req)
 {
 	char *keyring_name;
-	GkrKeyring *keyring;
+	GkrKeyring *keyring = NULL;
 	GnomeKeyringOpCode opcode;
 	GnomeKeyringResult res;
 	
@@ -1186,8 +1046,8 @@ op_change_keyring_password (GkrBuffer *packet, GkrBuffer *result,
 	
  out:
 	g_free (keyring_name);
-	gnome_keyring_free_password (original);
-	gnome_keyring_free_password (password);
+	gkr_secure_strfree (original);
+	gkr_secure_strfree (password);
 	
 	return TRUE;
 }
@@ -1270,8 +1130,7 @@ op_create_item (GkrBuffer *packet, GkrBuffer *result,
 		return FALSE;
 	}
 
-	if ((type & GNOME_KEYRING_ITEM_TYPE_MASK) >= GNOME_KEYRING_ITEM_LAST_TYPE ||
-	    display_name == NULL || secret == NULL) {
+	if (display_name == NULL || secret == NULL) {
 		res = GNOME_KEYRING_RESULT_BAD_ARGUMENTS;
 		goto out;
 	}
@@ -1301,7 +1160,7 @@ op_create_item (GkrBuffer *packet, GkrBuffer *result,
 	}
 	
 	if (update_if_exists) {
-		item = gkr_keyring_find_item (keyring, type, keyring->locked ? hashed : attributes);
+		item = gkr_keyring_find_item (keyring, type, keyring->locked ? hashed : attributes, TRUE);
 		if (item) {
 			/* Make sure we have access to the previous item */
 			if (!request_item_access (req, item, GNOME_KEYRING_ACCESS_WRITE, TRUE))
@@ -1319,8 +1178,8 @@ op_create_item (GkrBuffer *packet, GkrBuffer *result,
 
 	g_free (item->display_name);
 	item->display_name = g_strdup (display_name);
-	gnome_keyring_free_password (item->secret);
-	item->secret = gnome_keyring_memory_strdup (secret);
+	gkr_secure_strfree (item->secret);
+	item->secret = gkr_secure_strdup (secret);
 	gnome_keyring_attribute_list_free (item->attributes);
 	item->attributes = gnome_keyring_attribute_list_copy (attributes);
 	
@@ -1335,7 +1194,7 @@ op_create_item (GkrBuffer *packet, GkrBuffer *result,
  out:	
 	g_free (keyring_name);
 	g_free (display_name);
-	gnome_keyring_free_password (secret);
+	gkr_secure_strfree (secret);
 	gnome_keyring_attribute_list_free (hashed);
 	gnome_keyring_attribute_list_free (attributes);
 	
@@ -1573,8 +1432,8 @@ op_set_item_info (GkrBuffer *packet, GkrBuffer *result,
 			item->display_name = g_strdup (item_name);
 		}
 		if (secret != NULL) {
-			gnome_keyring_free_password (item->secret);
-			item->secret = gnome_keyring_memory_strdup (secret);
+			gkr_secure_strfree (item->secret);
+			item->secret = gkr_secure_strdup (secret);
 		}
 
 		if (item->keyring)
@@ -1583,7 +1442,7 @@ op_set_item_info (GkrBuffer *packet, GkrBuffer *result,
 
 	g_free (keyring_name);
 	g_free (item_name);
-	gnome_keyring_free_password (secret);
+	gkr_secure_strfree (secret);
 	return TRUE;
 }
 
@@ -1591,29 +1450,21 @@ static gboolean
 op_set_daemon_display (GkrBuffer *packet, GkrBuffer *result,
                        GkrKeyringRequest *req)
 {
-       char *display;
-       GnomeKeyringOpCode opcode;
+	char *display;
+	GnomeKeyringOpCode opcode;
 
-       if (!gkr_proto_decode_op_string (packet, &opcode, &display))
-               return FALSE;
+	if (!gkr_proto_decode_op_string (packet, &opcode, &display))
+		return FALSE;
 
-       if ( display == NULL ) {
-               gkr_buffer_add_uint32 (result, GNOME_KEYRING_RESULT_DENIED);
-               goto out;
-       }
+	if (display == NULL) {
+		gkr_buffer_add_uint32 (result, GNOME_KEYRING_RESULT_DENIED);
+	} else {
+		g_setenv ("DISPLAY", display, FALSE);
+		gkr_buffer_add_uint32 (result, GNOME_KEYRING_RESULT_OK);
+	}
 
-       if (gkr_ask_daemon_get_display () == NULL && (g_strrstr (display, ":") != NULL)) {
-               gkr_ask_daemon_set_display (display);
-       } else {
-               gkr_buffer_add_uint32 (result, GNOME_KEYRING_RESULT_DENIED);
-               goto out;
-       }
-
-       gkr_buffer_add_uint32 (result, GNOME_KEYRING_RESULT_OK);
-
-out:
-    g_free (display);
-       return TRUE;
+	g_free (display);
+	return TRUE;
 }
 
 static gboolean
@@ -1773,7 +1624,7 @@ op_find (GkrBuffer *packet, GkrBuffer *result, GkrKeyringRequest *req)
 
 	/* Need at least one attribute to match on */
 	if (ctx.attributes->len > 0) {
-		ctx.hashed = gkr_keyring_item_attributes_hash (ctx.attributes);
+		ctx.hashed = gkr_attribute_list_hash (ctx.attributes);
 		ctx.nfound = 0;
 		ctx.req = req;
 		ctx.items = NULL;
@@ -1826,6 +1677,45 @@ op_find (GkrBuffer *packet, GkrBuffer *result, GkrKeyringRequest *req)
 	return return_val;
 }
 
+static gboolean
+op_prepare_daemon_environment (GkrBuffer *packet, GkrBuffer *result, GkrKeyringRequest *req)
+{
+	const gchar **daemonenv;
+	gchar **environment, **e;
+	gchar *x;
+
+	if (!gkr_proto_decode_prepare_environment (packet, &environment))
+		return FALSE;
+
+	/* Accept environment from outside */
+	for (e = environment; *e; ++e) {
+		x = strchr (*e, '=');
+		if (x) {
+			*(x++) = 0;
+			
+			/* We're only interested in these guys */
+			if (g_str_equal (*e, "DISPLAY")) 
+				g_setenv ("DISPLAY", x, FALSE);
+			else if (g_str_equal (*e, "DBUS_SESSION_BUS_ADDRESS"))
+				g_setenv ("DBUS_SESSION_BUS_ADDRESS", x, FALSE);
+			else if (g_str_equal (*e, "XAUTHORITY"))
+				g_setenv ("XAUTHORITY", x, FALSE);
+			else if (g_str_equal (*e, "XDG_SESSION_COOKIE"))
+				g_setenv ("XDG_SESSION_COOKIE", x, FALSE);
+		}
+	}
+	
+	g_strfreev (environment);
+
+	gkr_buffer_add_uint32 (result, GNOME_KEYRING_RESULT_OK);
+
+	daemonenv = gkr_daemon_util_get_environment ();
+	g_return_val_if_fail (daemonenv, FALSE);
+	
+	gkr_buffer_add_stringv (result, daemonenv);
+	return TRUE;
+}
+
 GkrDaemonOperation keyring_ops[] = {
 	op_lock_all, 			/* LOCK_ALL */
 	op_set_default_keyring, 	/* SET_DEFAULT_KEYRING */
@@ -1850,4 +1740,5 @@ GkrDaemonOperation keyring_ops[] = {
 	op_change_keyring_password,     /* CHANGE_KEYRING_PASSWORD */
  	op_set_daemon_display,          /* SET_DAEMON_DISPLAY */
 	op_get_item_info,               /* GET_ITEM_INFO_PARTIAL */
+	op_prepare_daemon_environment,	/* PREPARE_DAEMON_ENVIRONMENT */
 };
